# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aimmo', '0005_auto_20160808_1545'),
    ]

    operations = [
        migrations.AddField(
            model_name='game',
            name='static_data',
            field=models.TextField(null=True, blank=True),
        ),
    ]



import docker
import platform
import re
from shell_api import (run_command, create_test_bin, BASE_DIR)
import json
import os


def create_docker_client(use_raw_env=False, minikube=None):
    """
    Create a docker client using the python SDK.

    :param raw_env_settings: String that is returned by the 'minikube docker-env' command.
    :return:
    """
    if use_raw_env:
        raw_env_settings = run_command([minikube, 'docker-env', '--shell="bash"'], True)
        matches = re.finditer(r'^export (.+)="(.+)"$', raw_env_settings, re.MULTILINE)
        env_variables = dict([(m.group(1), m.group(2)) for m in matches])
        
    else:
        # VM driver is set
        env_variables = os.environ

    env_variables['DOCKER_BUILDKIT'] = "1"
    return docker.from_env(
            environment=env_variables,
            version='auto',
            )


def build_docker_images(minikube=None, build_target=None):
    """
    Find environment settings and builds docker images for each directory.

    :param minikube: Executable command to run in terminal.
    """
    print('Building docker images')
    if minikube:
        client = create_docker_client(use_raw_env=True, minikube=minikube)
    else:
        client = create_docker_client(use_raw_env=False, minikube=minikube)

    directories = ('aimmo-game', 'aimmo-game-creator', 'aimmo-game-worker')
    for dir in directories:
        path = os.path.join(BASE_DIR, dir)
        tag = 'ocadotechnology/%s:test' % dir
        print("Building %s..." % tag)
        client.images.build(
            path=path,
            tag=tag,
            encoding='gzip',
            target=build_target
        )


def delete_containers():
    """Delete any containers starting with 'aimmo'."""
    client = docker.from_env(version='auto')

    containers = [container for container in client.containers.list(all=True) if container.name.startswith('aimmo')]
    for container in containers:
        container.remove(force=True)


def start_game_creator():
    """Start an aimmo-game-creator docker container."""
    os_name = platform.system()
    client = docker.from_env(version='auto')
    template = {
        'detach': True,
        'tty': True,
        'environment': {
            'FLASK_ENV': 'development',
            'WORKER_MANAGER': 'local'
        },
        'volumes': {
            '/var/run/docker.sock': {'bind': '/var/run/docker.sock', 'mode': 'rw'}
        }
    }
    if os_name == 'Linux':
        template['environment']['LOCALHOST_IP'] = '127.0.0.1'
        template['network_mode'] = 'host'
    else:
        template['environment']['LOCALHOST_IP'] = 'host.docker.internal'

    template['environment']['CONTAINER_TEMPLATE'] = json.dumps(template)
    kwargs = template.copy()
    client.containers.run(
        name='aimmo-game-creator',
        image='ocadotechnology/aimmo-game-creator:test',
        **kwargs
    )



from django.contrib import admin
from models import Avatar, Game


class GameDataAdmin(admin.ModelAdmin):
    list_display = ['id', 'name']
    
    readonly_fields = ['players']

    def players(self, obj):
        return "\n".join([u.first_name for u in obj.can_play.all()])


class AvatarDataAdmin(admin.ModelAdmin):
    list_display = ['id', 'owner_name', 'game_name']


    def owner_name(self, obj):
        return obj.owner

    def game_name(self, obj):
        return obj.game


admin.site.register(Game, GameDataAdmin)
admin.site.register(Avatar, AvatarDataAdmin)



from django import template

register = template.Library()


@register.inclusion_tag('players/dropdown.html', takes_context=True)
def game_dropdown_list(context, base_url):
    return get_user_playable_games(context, base_url)


def get_user_playable_games(context, base_url):
    user = context.request.user
    return {
        'base_url': base_url,
        'open_play_games': user.playable_games.all() if user.is_authenticated() else None
    }



class Location(object):
    def __init__(self, x, y):
        if not isinstance(x, int) or not isinstance(y, int):
            raise TypeError("Location needs to be an integer!")

        self.x = x
        self.y = y

    def __add__(self, direction):
        return Location(self.x + direction.x, self.y + direction.y)

    def __sub__(self, direction):
        return Location(self.x - direction.x, self.y - direction.y)

    def __repr__(self):
        return 'Location({}, {})'.format(self.x, self.y)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.x, self.y))

    def serialize(self):
        return {'x': self.x, 'y': self.y}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aimmo', '0003_auto_20160802_1418'),
    ]

    operations = [
        migrations.AddField(
            model_name='game',
            name='obstacle_ratio',
            field=models.FloatField(default=0.1),
        ),
        migrations.AddField(
            model_name='game',
            name='pickup_spawn_chance',
            field=models.FloatField(default=0.02),
        ),
        migrations.AddField(
            model_name='game',
            name='score_despawn_chance',
            field=models.FloatField(default=0.02),
        ),
        migrations.AddField(
            model_name='game',
            name='start_height',
            field=models.IntegerField(default=11),
        ),
        migrations.AddField(
            model_name='game',
            name='start_width',
            field=models.IntegerField(default=11),
        ),
        migrations.AddField(
            model_name='game',
            name='target_num_cells_per_avatar',
            field=models.FloatField(default=16),
        ),
        migrations.AddField(
            model_name='game',
            name='target_num_pickups_per_avatar',
            field=models.FloatField(default=0.5),
        ),
        migrations.AddField(
            model_name='game',
            name='target_num_score_locations_per_avatar',
            field=models.FloatField(default=0.5),
        ),
    ]



import mock
from unittest import TestCase
from requests import Response
from simulation.worker import Worker

DEFAULT_RESPONSE_CONTENT = b'{"action": "test_action",' \
                           b'"log": "test_log",' \
                           b'"avatar_updated": "True"}'

MISSING_KEY_RESPONSE_CONTENT = b'{"corruptedKey": "test_action",' \
                               b'"log": "test_log",' \
                               b'"avatar_updated": "True"}'


def construct_test_response(status_code=200,
                            response_content=DEFAULT_RESPONSE_CONTENT):
    post_response = Response()
    post_response.status_code = status_code
    post_response._content = response_content
    return post_response


class TestWorker(TestCase):
    def setUp(self):
        self.worker = Worker(worker_url='http://test')

    @mock.patch('simulation.worker.requests.post',
                return_value=construct_test_response())
    def test_fetch_data_fetches_correct_response(self, mocked_post):
        self.worker.fetch_data(state_view={})

        mocked_post.assert_called_once()
        self.assertEqual(self.worker.serialized_action, 'test_action')
        self.assertEqual(self.worker.log, 'test_log')
        self.assertEqual(self.worker.has_code_updated, 'True')

    def test_setting_defaults_works_correctly(self):
        self.worker.log = 'test_log_fake'
        self.worker.serialized_action = 'test_action_fake'
        self.worker.has_code_updated = 'test_avatar_updated_fake'

        self.worker._set_defaults()

        self.assertIsNone(self.worker.serialized_action)
        self.assertIsNone(self.worker.log)
        self.assertFalse(self.worker.has_code_updated)

    @mock.patch('simulation.worker.requests.post',
                return_value=construct_test_response(status_code=500))
    @mock.patch.object(target=Worker, attribute='_set_defaults')
    def test_fetch_data_cannot_connect_to_worker(self, mocked_set_defaults,
                                                 mocked_post):
        self.worker.fetch_data(state_view={})

        mocked_post.assert_called_once()
        mocked_set_defaults.assert_called_once()

    @mock.patch('simulation.worker.requests.post',
                return_value=construct_test_response(response_content=MISSING_KEY_RESPONSE_CONTENT))
    @mock.patch.object(target=Worker, attribute='_set_defaults')
    def test_missing_key_in_worker_data(self, mocked_set_defaults,
                                        mocked_post):
        self.worker.fetch_data(state_view={})

        mocked_post.assert_called_once()
        mocked_set_defaults.assert_called_once()



class Avatar:
    def next_turn(self, world_state, avatar_state):
        self.world_state = world_state
        self.avatar_state = avatar_state

        if world_state.get_cell(avatar_state.location).generates_score:
            return WaitAction()

        possible_directions = self.get_possible_directions()
        directions_to_emphasise = [d for d in possible_directions if self.is_towards(d, self.get_closest_score_location())]
        return MoveAction(random.choice(possible_directions + (directions_to_emphasise * 5)))

    def is_towards(self, direction, location):
        if location:
            return self.distance_between(self.avatar_state.location, location) > \
                self.distance_between(self.avatar_state.location + direction, location)
        else:
            return False

    def distance_between(self, a, b):
        return abs(a.x - b.x) + abs(a.y - b.y)

    def get_closest_score_location(self):
        score_cells = list(self.world_state.score_cells())
        if score_cells:
            return min(score_cells, key=lambda cell: self.distance_between(cell.location, self.avatar_state.location)).location
        else:
            return None

    def get_possible_directions(self):
        directions = (direction.EAST, direction.SOUTH, direction.WEST, direction.NORTH)
        return [d for d in directions if self.world_state.can_move_to(self.avatar_state.location + d)]



import cPickle as pickle
import logging
import os
import json

from django.contrib.auth.decorators import login_required
from django.core.exceptions import ValidationError
from django.http import HttpResponse, JsonResponse, Http404, HttpResponseForbidden
from django.shortcuts import redirect, render, get_object_or_404
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods
from django.views.generic import TemplateView
from django.middleware.csrf import get_token

from models import Avatar, Game, LevelAttempt
from aimmo import forms, game_renderer
from app_settings import get_users_for_new_game, preview_user_required
from exceptions import UserCannotPlayGameException

LOGGER = logging.getLogger(__name__)


def _post_code_success_response(message):
    return _create_response("SUCCESS", message)


def _create_response(status, message):
    response = {
        "status": status,
        "message": message
    }
    return JsonResponse(response)


@login_required
@preview_user_required
def code(request, id):
    if not request.user:
        return HttpResponseForbidden()
    game = get_object_or_404(Game, id=id)
    if not game.can_user_play(request.user):
        raise Http404
    try:
        avatar = game.avatar_set.get(owner=request.user)
    except Avatar.DoesNotExist:
        initial_code_file_name = os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            'avatar_examples/simple_avatar.py',
        )
        with open(initial_code_file_name) as initial_code_file:
            initial_code = initial_code_file.read()
        avatar = Avatar.objects.create(owner=request.user, code=initial_code,
                                       game_id=id)
    if request.method == 'POST':
        avatar.code = request.POST['code']
        avatar.save()
        return HttpResponse(status=200)
    else:
        return JsonResponse({'code': avatar.code})


def list_games(request):
    response = {
        game.pk:
            {
                'name': game.name,
                'settings': json.dumps(game.settings_as_dict()),
            } for game in Game.objects.exclude_inactive()
    }
    return JsonResponse(response)


def get_game(request, id):
    game = get_object_or_404(Game, id=id)
    response = {
        'main': {
            'parameters': [],
            'main_avatar': None,
            'users': [],
        }
    }
    for avatar in game.avatar_set.all():
        if avatar.owner_id == game.main_user_id:
            response['main']['main_avatar'] = avatar.id
        response['main']['users'].append({
            'id': avatar.id,
            'code': avatar.code,
        })
    return JsonResponse(response)


def connection_parameters(request, game_id):
    """
    An API view which returns the correct connection settings required
    to run the game in different environments. These values will change
    depending on where the project is started (ie. local, etc).
    :param request: Django request object.
    :param game_id: Integer with the ID of the game.
    :return: JsonResponse object with the contents.
    """
    env_connection_settings = game_renderer.get_environment_connection_settings(game_id)

    avatar_id, response = get_avatar_id(request, game_id)

    if avatar_id:
        env_connection_settings.update({'avatar_id': avatar_id})
        return JsonResponse(env_connection_settings)
    else:
        return response


@csrf_exempt
@require_http_methods(['POST'])
def mark_game_complete(request, id):
    game = get_object_or_404(Game, id=id)
    game.completed = True
    game.static_data = request.body
    game.save()
    return HttpResponse('Done!')


class ProgramView(TemplateView):
    template_name = 'players/program.html'

    def get_context_data(self, **kwargs):
        context = super(ProgramView, self).get_context_data(**kwargs)
        game = get_object_or_404(Game, id=self.kwargs['id'])
        if not game.can_user_play(self.request.user):
            raise Http404
        context['game_id'] = int(self.kwargs['id'])
        return context


def watch_game(request, id):
    game = get_object_or_404(Game, id=id)
    if not game.can_user_play(request.user):
        raise Http404
    return game_renderer.render_game(request, game)


def watch_level(request, num):
    try:
        game = Game.objects.get(levelattempt__user=request.user, levelattempt__level_number=num)
    except Game.DoesNotExist:
        LOGGER.debug('Adding level')
        game = _add_and_return_level(num, request.user)
    LOGGER.debug('Displaying game with id %s', game.id)
    return game_renderer.render_game(request, game)


def _add_and_return_level(num, user):
    game = Game(generator='Level' + num, name='Level ' + num, public=False, main_user=user)
    try:
        game.save()
    except ValidationError as e:
        LOGGER.warn(e)
        raise Http404
    game.can_play = [user]
    game.save()
    level_attempt = LevelAttempt(game=game, user=user, level_number=num)
    level_attempt.save()
    return game


@login_required
@preview_user_required
def add_game(request):
    playable_games = request.user.playable_games.all()

    if request.method == 'POST':
        form = forms.AddGameForm(playable_games, data=request.POST)
        if form.is_valid():
            game = form.save(commit=False)
            game.generator = 'Main'
            game.owner = request.user
            game.main_user = request.user
            game.save()
            users = get_users_for_new_game(request)
            if users is not None:
                game.can_play.add(*users)
            return redirect('aimmo/play', id=game.id)
    else:
        form = forms.AddGameForm(playable_games)
    return render(request, 'players/add_game.html', {'form': form})


def current_avatar_in_game(request, game_id):
    avatar_id, response = get_avatar_id(request, game_id)

    if avatar_id:
        return JsonResponse({'current_avatar_id': avatar_id})
    else:
        return response


def get_avatar_id(request, game_id):
    avatar_id = None
    response = None

    try:
        avatar_id = game_renderer.get_avatar_id_from_user(user=request.user, game_id=game_id)
    except UserCannotPlayGameException:
        LOGGER.warning('HTTP 401 returned. User {} unauthorised to play.'.format(request.user.id))
        response = HttpResponse('User unauthorized to play', status=401)
    except Avatar.DoesNotExist:
        LOGGER.warning('Avatar does not exist for user {} in game {}'.format(request.user.id, game_id))
        response = HttpResponse('Avatar does not exist for this user', status=404)
    except Exception as e:
        LOGGER.error('Unknown error occurred while getting connection parameters!')
        LOGGER.error(e)
        response = HttpResponse('Unknown error occurred when getting the current avatar', status=500)

    return avatar_id, response


def csrfToken(request):
    if request.method == 'GET':
        token = get_token(request)
        return JsonResponse({'csrfToken': token})
    else:
        return HttpResponse(status=405)



from .spawn_location_finder import SpawnLocationFinder
from .map_updaters import ScoreLocationUpdater, MapContext, PickupUpdater, MapExpander
from .pickup_applier import PickupApplier

__all__ = [
    'SpawnLocationFinder',
    'ScoreLocationUpdater',
    'MapContext',
    'PickupUpdater',
    'MapExpander',
    'PickupApplier'
]



import asyncio
import concurrent.futures
import logging
import threading
import time

from prometheus_client import Histogram

from metrics import GAME_TURN_TIME
from simulation.avatar.avatar_manager import AvatarManager
from simulation.django_communicator import DjangoCommunicator
from simulation.simulation_runner import ConcurrentSimulationRunner

LOGGER = logging.getLogger(__name__)

TURN_TIME = 2


class GameRunner:
    def __init__(self, worker_manager_class, game_state_generator, django_api_url, port):
        super(GameRunner, self).__init__()

        self.worker_manager = worker_manager_class(port=port)
        self.game_state = game_state_generator(AvatarManager())
        self.communicator = DjangoCommunicator(django_api_url=django_api_url,
                                               completion_url=django_api_url + 'complete/')
        self.simulation_runner = ConcurrentSimulationRunner(communicator=self.communicator,
                                                            game_state=self.game_state)
        self._end_turn_callback = lambda: None

    def set_end_turn_callback(self, callback_method):
        self._end_turn_callback = callback_method

    def get_users_to_add(self, game_metadata):
        def player_is_new(_player):
            return _player['id'] not in self.worker_manager.player_id_to_worker.keys()

        return [player['id'] for player in game_metadata['users'] if player_is_new(player)]

    def get_users_to_delete(self, game_metadata):
        def player_in_worker_manager_but_not_metadata(pid):
            return pid not in [player['id'] for player in game_metadata['users']]

        return [player_id for player_id in self.worker_manager.player_id_to_worker.keys()
                if player_in_worker_manager_but_not_metadata(player_id)]

    def update_main_user(self, game_metadata):
        self.game_state.main_avatar_id = game_metadata['main_avatar']

    def update_workers(self):
        game_metadata = self.communicator.get_game_metadata()['main']

        users_to_add = self.get_users_to_add(game_metadata)
        users_to_delete = self.get_users_to_delete(game_metadata)

        self.worker_manager.add_workers(users_to_add)
        self.worker_manager.delete_workers(users_to_delete)
        self.simulation_runner.add_avatars(users_to_add)
        self.simulation_runner.delete_avatars(users_to_delete)
        self.worker_manager.update_worker_codes(game_metadata['users'])

        self.update_main_user(game_metadata)
        self.worker_manager.fetch_all_worker_data(self.game_state.get_serialized_game_states_for_workers())

    async def update_simulation(self, player_id_to_serialized_actions):
        await self.simulation_runner.run_single_turn(player_id_to_serialized_actions)
        await self._end_turn_callback()

    async def update(self):
        with GAME_TURN_TIME():
            self.update_workers()
            await self.update_simulation(self.worker_manager.get_player_id_to_serialized_actions())
            self.worker_manager.clear_logs()

    async def run(self):
        while True:
            await self.update()



class Avatar:
    def next_turn(self, world_state, avatar_state):
        self.avatar_state = avatar_state
        self.location = self.avatar_state.location

        directions = (direction.EAST, direction.SOUTH, direction.WEST, direction.NORTH)
        direction_of_other_avatar = next((d for d in directions if world_state.is_visible(self.location + d) and world_state.get_cell(self.location + d).avatar), None)
        if direction_of_other_avatar:
            return AttackAction(direction_of_other_avatar)

        direction_to_other_player = self.direction_to(next(cell.location for cell in world_state.all_cells() if cell.avatar and cell.location != avatar_state.location))
        if direction_to_other_player:
            return MoveAction(random.choice(directions + ((direction_to_other_player,) * 10)))
        return MoveAction(random.choice(directions))

    def direction_to(self, location):
        vector_to = location - self.location
        if vector_to.x != 0:
            return direction.Direction(1 if vector_to.x > 0 else -1, 0)
        if vector_to.y != 0:
            return direction.Direction(0, 1 if vector_to.y > 0 else -1)
        return None



import logging

import requests
LOGGER = logging.getLogger(__name__)


class Worker(object):
    def __init__(self, worker_url):
        self.url = worker_url
        self.log = None
        self.code = None
        self.serialized_action = None
        self.has_code_updated = False

    def _set_defaults(self):
        self.log = None
        self.serialized_action = None
        self.has_code_updated = False

    def fetch_data(self, state_view):
        try:
            code_and_options = {
                'code': self.code,
                'options': {},
                'state': None,
            }
            data = {**state_view, **code_and_options}
            response = requests.post(self.url, json=data)
            response.raise_for_status()
            data = response.json()
            self.serialized_action = data['action']
            self.log = data['log']
            self.has_code_updated = data['avatar_updated']
        except requests.exceptions.ConnectionError:
            LOGGER.info('Could not connect to worker, probably not ready yet')
            self._set_defaults()
        except KeyError as e:
            LOGGER.error('Missing key in data from worker: {}'.format(e))
            self._set_defaults()
        except Exception as e:
            LOGGER.exception('Unknown error while fetching turn data.')
            LOGGER.exception(e)
            self._set_defaults()



import argparse
import os
import uuid

from django.core.management import BaseCommand
from django.test import RequestFactory
from importlib import import_module
from django.conf import settings
from django.contrib.auth.models import User
from aimmo.views import code as code_view


def _nth_dirname(path, n):
    for _ in xrange(n):
        path = os.path.dirname(path)
    return path

_PLAYERS_DIRECTORY = _nth_dirname(__file__, 3)
_AVATAR_CODES_DIRECTORY = os.path.join(_PLAYERS_DIRECTORY, 'avatar_examples')

# Code file listing


def _strip_prefix(prefix, string):
    if string.startswith(prefix):
        return string[len(prefix):]


def _get_available_code_files(base_directory):
    for dirpath, dirnames, filenames in os.walk(base_directory):
        for f in filenames:
            if not f.startswith('_') and f.endswith('.py'):
                parent_dir = _strip_prefix(base_directory, dirpath)
                yield os.path.join(parent_dir, f)

_AVATAR_CODES = list(_get_available_code_files(_AVATAR_CODES_DIRECTORY))


# Code file loading
def _load_code_file(filename):
    if not filename.endswith('.py'):
        filename = filename + '.py'

    filepath = os.path.join(_AVATAR_CODES_DIRECTORY, filename)

    with open(filepath) as f:
        return f.read()


class LoadCodeAction(argparse.Action):

    def __init__(self, option_strings, dest, **kwargs):
        super(LoadCodeAction, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string):
        values = _load_code_file(values)
        setattr(namespace, self.dest, values)


class Command(BaseCommand):
    # Show this when the user types help
    help = "Generate users for the game"

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.engine = import_module(settings.SESSION_ENGINE)
        self.request_factory = RequestFactory()

    def add_arguments(self, parser):
        parser.add_argument('num-users', type=int,
                            help='Number of users to create')
        parser.add_argument('avatar-code', choices=_AVATAR_CODES,
                            action=LoadCodeAction,
                            help='The code to use for the avatar.')
        parser.add_argument('game-id', type=int)

    # A command must define handle()
    def handle(self, *args, **options):
        num_users = options['num-users']
        code = options['avatar-code']
        game_id = options['game-id']

        for _ in xrange(num_users):
            random_string = str(uuid.uuid4())[:8]
            username = 'zombie-%s' % random_string
            password = '123'
            user = self.create_user(username, password)
            self.post_code(user, code, game_id)

    def create_user(self, username, password):
        user = User.objects.create_user(username, 'user@example.com', password)
        self.stdout.write('Created user %s with password: %s' % (username, password))
        return user

    def post_code(self, user, player_code, game_id):
        request = self.request_factory.post('/any_path', data={'code': player_code})
        session_key = None
        request.session = self.engine.SessionStore(session_key)
        request.user = user

        response = code_view(request, game_id)
        if response.status_code == 200:
            self.stdout.write('Posted code for player %s' % user)
        else:
            raise Exception('Failed to submit code for player %s' % user)



import logging
import os
import time

from kubernetes import client
from kubernetes import config

from .worker_manager import WorkerManager

LOGGER = logging.getLogger(__name__)
# Default here to stop import errors if imported when running locally
K8S_NAMESPACE = os.environ.get('K8S_NAMESPACE', '')


class KubernetesWorkerManager(WorkerManager):
    """Kubernetes worker manager."""

    def __init__(self, *args, **kwargs):
        config.load_incluster_config()
        self.api = client.CoreV1Api()
        self.game_id = os.environ['GAME_ID']
        self.game_url = os.environ['GAME_URL']
        self.pod_name = os.environ['POD_NAME']
        super(KubernetesWorkerManager, self).__init__(*args, **kwargs)

    @staticmethod
    def _create_a_label_selector_from_labels(label_list):
        return ','.join(label_list)

    def _make_owner_references(self):
        try:
            return [client.V1OwnerReference(
                api_version="v1",
                block_owner_deletion=True,
                controller=True,
                kind="Pod",
                name=self.pod_name,
                uid=self._get_game_uid()
            )]
        except IndexError:
            # Couldn't find the current pod
            return []

    def _make_container(self, player_id):
        return client.V1Container(
            env=[client.V1EnvVar(
                name='DATA_URL',
                value='%s/player/%d' % (self.game_url, player_id)),
                client.V1EnvVar(
                name='PORT',
                value='5000')],
            name='aimmo-game-worker',
            image='ocadotechnology/aimmo-game-worker:%s' % os.environ.get('IMAGE_SUFFIX', 'latest'),
            ports=[client.V1ContainerPort(
                container_port=5000,
                protocol='TCP')],
            resources=client.V1ResourceRequirements(
                limits={'cpu': '10m', 'memory': '64Mi'},
                requests={'cpu': '6m', 'memory': '32Mi'}),
            security_context=client.V1SecurityContext(
                capabilities=client.V1Capabilities(
                    drop=['all'],
                    add=['NET_BIND_SERVICE']))
        )

    def make_pod(self, player_id):
        pod_manifest = client.V1PodSpec(containers=[self._make_container(player_id)], service_account_name='worker')

        metadata = client.V1ObjectMeta(
            labels={
                'app': 'aimmo-game-worker',
                'game': self.game_id,
                'player': str(player_id)},
            generate_name='aimmo-%s-worker-%s-' % (self.game_id, player_id),
            owner_references=self._make_owner_references()
        )
        return client.V1Pod(metadata=metadata, spec=pod_manifest)

    def _get_game_uid(self):
        pod_list = self.api.list_namespaced_pod(namespace=K8S_NAMESPACE,
                                                field_selector=f'metadata.name={self.pod_name}')
        pod_metadata = pod_list.items[0].metadata
        return pod_metadata.uid

    def _wait_for_pod_creation(self, pod_name, player_id):

        for _ in range(90):
            pod = self.api.read_namespaced_pod(pod_name, K8S_NAMESPACE)
            LOGGER.info('Pod status: {}'.format(pod.status))
            if pod.status.phase == 'Running':
                return pod

            time.sleep(1)

        raise EnvironmentError('Could not start worker %s.' % player_id)

    def create_worker(self, player_id):
        pod_obj = self.make_pod(player_id)
        LOGGER.info('Making new worker pod: {}'.format(pod_obj.metadata.name))
        pod = self.api.create_namespaced_pod(namespace=K8S_NAMESPACE, body=pod_obj)
        pod_name = pod.metadata.name
        pod = self._wait_for_pod_creation(pod_name, player_id)

        worker_url = 'http://%s:5000' % pod.status.pod_ip
        LOGGER.info('Worker ip: {}'.format(pod.status.pod_ip))
        LOGGER.info('Worker started for %s, listening at %s', player_id, worker_url)
        return worker_url

    def remove_worker(self, player_id):
        app_label = 'app=aimmo-game-worker'
        game_label = 'game={}'.format(self.game_id)
        player_label = 'player={}'.format(player_id)
        label_selector = self._create_a_label_selector_from_labels([app_label,
                                                                    game_label,
                                                                    player_label])
        pods = self.api.list_namespaced_pod(namespace=K8S_NAMESPACE,
                                            label_selector=label_selector)

        for pod in pods.items:
            LOGGER.info('Deleting pod: {}'.format(pod.metadata.name))
            self.api.delete_namespaced_pod(pod.metadata.name, K8S_NAMESPACE, client.V1DeleteOptions())



from threading import RLock

from simulation.pickups.pickup_types import serialize_pickups


class GameState:
    """
    Encapsulates the entire game state, including avatars, their code, and the world.
    """

    def __init__(self, world_map, avatar_manager, completion_check_callback=lambda: None):
        self.world_map = world_map
        self.avatar_manager = avatar_manager
        self.main_avatar_id = None
        self._lock = RLock()

    def get_main_avatar(self):
        with self._lock:
            return self.avatar_manager.avatars_by_id[self.main_avatar_id]

    def serialize(self):
        return {
            'era': "less_flat",
            'southWestCorner': self.world_map.get_serialized_south_west_corner(),
            'northEastCorner': self.world_map.get_serialized_north_east_corner(),
            'players': self.avatar_manager.serialize_players(),
            'pickups': serialize_pickups(self.world_map),
            'scoreLocations': (self.world_map.serialize_score_location()),
            'obstacles': self.world_map.serialize_obstacles()
        }

    def serialize_for_worker(self, avatar_wrapper):
        with self._lock:
            return {
                'avatar_state': avatar_wrapper.serialize(),
                'world_map': {
                    'cells': [cell.serialize() for cell in self.world_map.all_cells()]
                }
            }

    def get_serialized_game_states_for_workers(self):
        with self._lock:
            return {player_id: self.serialize_for_worker(avatar_wrapper) for player_id, avatar_wrapper
                    in self.avatar_manager.avatars_by_id.items()}



from abc import ABCMeta, abstractmethod, abstractproperty
from functools import reduce
from logging import getLogger
from typing import TYPE_CHECKING, List

from simulation.pickups.conditions import TurnState, avatar_on_cell
from simulation.pickups.effects import (DamageBoostEffect, HealthEffect,
                                        InvulnerabilityEffect)

if TYPE_CHECKING:
    from simulation.game_state import GameState
    from simulation.avatar.avatar_wrapper import AvatarWrapper

LOGGER = getLogger(__name__)


class _Pickup(object):
    __metaclass__ = ABCMeta

    def __init__(self, cell):
        self.cell = cell
        self.conditions = []
        self.effects = []
        self.targets = []

    def __str__(self):
        return self.__class__.__name__

    def delete(self):
        self.cell.pickup = None

    def conditions_met(self, game_state: 'GameState') -> 'bool':
        """Apply logical `AND` on all conditions, returns True is all conditions are met."""
        turn_state = TurnState(game_state, self.cell)
        return all([condition(turn_state) for condition in self.conditions])

    def apply(self):
        """
        Apply all effects in sequential order.

        Targets for effects can be a single object, or a list of objects. all
        targets must have an 'effect' attribute that is of type=set.
        """
        self.targets = self.get_targets()
        for effect in self.effects:
            for target in self.targets:
                effect(target)

        self.delete()

    @abstractmethod
    def serialize(self):
        raise NotImplementedError()

    @abstractmethod
    def get_targets(self) -> 'List[AvatarWrapper]':
        raise NotImplementedError()


class HealthPickup(_Pickup):
    def __init__(self, cell):
        super(HealthPickup, self).__init__(cell)
        self.conditions.append(avatar_on_cell)

        self.effects.append(HealthEffect)

    def get_targets(self):
        return [
            self.cell.avatar
        ]

    def __repr__(self):
        return 'HealthPickup(Location={})'.format(self.cell.location)

    def serialize(self):
        return {
            'type': 'health',
            'location': {
                    'x': self.cell.location.x,
                    'y': self.cell.location.y,
            }
        }


class InvulnerabilityPickup(_Pickup):
    def __init__(self, cell):
        super(InvulnerabilityPickup, self).__init__(cell)
        self.conditions.append(avatar_on_cell)

        self.effects.append(InvulnerabilityEffect)

    def get_targets(self):
        return [
            self.cell.avatar
        ]

    def __repr__(self):
        return 'InvulnerabilityPickup(Location={})'.format(self.cell.location)

    def serialize(self):
        return {
            'type': 'invulnerability',
            'location': {
                    'x': self.cell.location.x,
                    'y': self.cell.location.y,
            }
        }


class DamageBoostPickup(_Pickup):
    def __init__(self, cell):
        super(DamageBoostPickup, self).__init__(cell)
        self.conditions.append(avatar_on_cell)

        self.effects.append(DamageBoostEffect)

    def get_targets(self):
        return [
            self.cell.avatar
        ]

    def __repr__(self):
        return 'DamageBoostPickup(Location={})'.format(self.cell.location)

    def serialize(self):
        return {
            'type': 'damage_boost',
            'location': {
                    'x': self.cell.location.x,
                    'y': self.cell.location.y,
            }
        }


def serialize_pickups(world_map):
    return [cell.pickup.serialize() for cell in world_map.pickup_cells()]


ALL_PICKUPS = (
    HealthPickup,
    InvulnerabilityPickup,
    DamageBoostPickup,
)



from simulation.pickups.pickup_types import (ALL_PICKUPS, DamageBoostPickup,
                                             HealthPickup,
                                             InvulnerabilityPickup,
                                             serialize_pickups)

__all__ = [
    'ALL_PICKUPS',
    'DamageBoostPickup',
    'InvulnerabilityPickup',
    'HealthPickup',
    'serialize_pickups'
]



from .kubernetes_worker_manager import KubernetesWorkerManager as _KWorkerManager
from .local_worker_manager import LocalWorkerManager as _LWorkerManager

WORKER_MANAGERS = {
    'local': _LWorkerManager,
    'kubernetes': _KWorkerManager,
}



from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from simulation.game_state import GameState
    from simulation.cell import Cell


class TurnState:
    """
    Data structure used by conditions.

    On a given turn, all conditions for a pickup get access to the same TurnState object.
    """

    def __init__(self, game_state: 'GameState', pickup_cell: 'Cell'):
        self.game_state = game_state
        self.pickup_cell = pickup_cell


def avatar_on_cell(turn_state: TurnState):
    """ Returns an expression that checks if an avatar is on a specified cell. """
    return turn_state.pickup_cell.avatar is not None



from __future__ import absolute_import

import asyncio
import unittest
from string import ascii_uppercase
from unittest.mock import patch

from simulation.avatar.avatar_appearance import AvatarAppearance
from simulation.game_state import GameState
from simulation.location import Location
from simulation.pickups import DamageBoostPickup
from simulation.simulation_runner import ConcurrentSimulationRunner
from simulation.world_map import WorldMap

from .dummy_avatar import (DeadDummy, DummyAvatar, DummyAvatarManager,
                           MoveEastDummy, MoveNorthDummy, MoveSouthDummy,
                           MoveWestDummy, WaitDummy)
from .maps import InfiniteMap, MockCell, MockPickup
from .mock_communicator import MockCommunicator

ORIGIN = Location(0, 0)

RIGHT_OF_ORIGIN = Location(1, 0)
FIVE_RIGHT_OF_ORIGIN = Location(5, 0)

ABOVE_ORIGIN = Location(0, 1)
FIVE_RIGHT_OF_ORIGIN_AND_ONE_ABOVE = Location(5, 1)

SETTINGS = {
            'TARGET_NUM_CELLS_PER_AVATAR': 0,
            'TARGET_NUM_PICKUPS_PER_AVATAR': 0,
            'TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR': 0,
            'SCORE_DESPAWN_CHANCE': 0,
            'PICKUP_SPAWN_CHANCE': 0,
        }


class MockGameState(GameState):
    def get_state_for(self, avatar):
        return self


class TestSimulationRunner(unittest.TestCase):
    """
        Key:
            > : Avatar moving eastward
            < : Avatar moving westward
            x : Avatar waiting / blocked
            o : Avatar successfully moved
            ! : Dead avatar (that should be waiting)
    """

    def _generate_grid(self, columns=2, rows=2):
        alphabet = iter(ascii_uppercase)
        grid = {Location(x, y): MockCell(Location(x, y), name=next(alphabet))
                for x in range(columns) for y in range(rows)}
        return grid

    def assertGridSize(self, world_map, expected_columns, expected_rows=None):
        if expected_rows is None:
            expected_rows = expected_columns
        self.assertEqual(world_map.num_rows, expected_rows)
        self.assertEqual(world_map.num_cols, expected_columns)
        self.assertEqual(world_map.num_cells, expected_rows*expected_columns)
        self.assertEqual(len(list(world_map.all_cells())), expected_rows*expected_columns)

    def construct_default_avatar_appearance(self):
        return AvatarAppearance("#000", "#ddd", "#777", "#fff")

    def construct_simulation_runner(self, avatars, locations):
        self.avatar_manager = DummyAvatarManager(avatars)
        self.avatar_manager.avatars_by_id = dict(enumerate(avatars))
        self.game_state = MockGameState(InfiniteMap(), self.avatar_manager)
        self.simulation_runner = ConcurrentSimulationRunner(game_state=self.game_state,
                                                            communicator=MockCommunicator())
        for index, location in enumerate(locations):
            self.simulation_runner.add_avatar(index, location)

    def assert_at(self, avatar, location):
        self.assertEqual(avatar.location, location)
        cell = self.game_state.world_map.get_cell(location)
        self.assertEqual(cell.avatar, avatar)

    def get_avatar(self, player_id):
        return self.avatar_manager.get_avatar(player_id)

    def run_turn(self):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.simulation_runner.run_turn(self.avatar_manager.avatars_by_id))

    def test_add_avatar(self):
        self.construct_simulation_runner([], [])

        self.simulation_runner.add_avatar(7)
        self.assertIn(7, self.game_state.avatar_manager.avatars_by_id)

        avatar = self.game_state.avatar_manager.avatars_by_id[7]
        self.assertEqual(avatar.location.x, 2)
        self.assertEqual(avatar.location.y, 2)

    def test_remove_avatar(self):
        self.construct_simulation_runner(
            [DummyAvatar, DummyAvatar],
            [Location(0, 0), Location(1, 1)]
        )

        avatar = self.get_avatar(1)
        avatar.marked = True

        self.simulation_runner.remove_avatar(0)

        self.assertNotIn(0, self.avatar_manager.avatars_by_id)
        self.assertEqual(self.game_state.world_map.get_cell(Location(0, 0)).avatar, None)

        self.assertTrue(self.avatar_manager.avatars_by_id[1].marked)
        self.assertTrue(self.game_state.world_map.get_cell(Location(1, 1)).avatar.marked)

    def test_remove_non_existent_avatar(self):
        self.construct_simulation_runner([], [])
        self.simulation_runner.remove_avatar(10)

    def test_updates_map_with_correct_num_avatars(self):
        self.construct_simulation_runner([], [])

        self.avatar_manager.add_avatar(1)
        self.simulation_runner.update_environment()
        self.assertEqual(len(self.simulation_runner.game_state.avatar_manager.avatars_by_id), 1)

        self.avatar_manager.add_avatar(2)
        self.avatar_manager.add_avatar(3)
        self.simulation_runner.update_environment()
        self.assertEqual(len(self.simulation_runner.game_state.avatar_manager.avatars_by_id), 3)

    def test_grid_expand(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings['TARGET_NUM_CELLS_PER_AVATAR'] = 5
        self.simulation_runner.game_state.world_map = WorldMap(self._generate_grid(), settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        print(self.simulation_runner.game_state.world_map)
        self.assertTrue(self.simulation_runner.game_state.world_map.is_on_map(Location(-1, -1)))
        self.assertTrue(self.simulation_runner.game_state.world_map.is_on_map(Location(-1, 2)))
        self.assertTrue(self.simulation_runner.game_state.world_map.is_on_map(Location(2, 2)))
        self.assertTrue(self.simulation_runner.game_state.world_map.is_on_map(Location(2, -1)))
        self.assertGridSize(self.simulation_runner.game_state.world_map, 4)

        self.simulation_runner.update(4, self.simulation_runner.game_state)
        self.assertGridSize(self.simulation_runner.game_state.world_map, 6)
        self.assertTrue(self.simulation_runner.game_state.world_map.is_on_map(Location(0, 3)))
        self.assertTrue(self.simulation_runner.game_state.world_map.is_on_map(Location(3, 0)))
        self.assertTrue(self.simulation_runner.game_state.world_map.is_on_map(Location(-2, 0)))
        self.assertTrue(self.simulation_runner.game_state.world_map.is_on_map(Location(0, -2)))

    def test_grid_doesnt_expand(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings['TARGET_NUM_CELLS_PER_AVATAR'] = 4
        self.simulation_runner.game_state.world_map = WorldMap(self._generate_grid(), settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertGridSize(self.simulation_runner.game_state.world_map, 2)

    def test_scores_removed(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings['SCORE_DESPAWN_CHANCE'] = 1
        grid = self._generate_grid()
        grid[Location(0, 1)].generates_score = True
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertEqual(len(list(self.simulation_runner.game_state.world_map.score_cells())), 0)

    def test_score_despawn_chance(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings['TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR'] = 0
        grid = self._generate_grid()
        grid[Location(0, 1)].generates_score = True
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertIn(grid[Location(0, 1)], self.simulation_runner.game_state.world_map.score_cells())
        self.assertEqual(len(list(self.simulation_runner.game_state.world_map.score_cells())), 1)

    def test_scores_added(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings['TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR'] = 1
        self.simulation_runner.game_state.world_map = WorldMap(self._generate_grid(), settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertEqual(len(list(self.simulation_runner.game_state.world_map.score_cells())), 1)

        self.simulation_runner.update(2, self.simulation_runner.game_state)
        self.assertEqual(len(list(self.simulation_runner.game_state.world_map.score_cells())), 2)

    def test_scores_applied(self):
        self.construct_simulation_runner([], [])
        grid = self._generate_grid()
        avatar = DummyAvatar()
        grid[Location(1, 1)].generates_score = True
        grid[Location(1, 1)].avatar = avatar
        self.simulation_runner.game_state.world_map = WorldMap(grid, SETTINGS)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertEqual(avatar.score, 1)

    def test_scores_not_added_when_at_target(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings['TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR'] = 1
        grid = self._generate_grid()
        grid[Location(0, 1)].generates_score = True
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertEqual(len(list(self.simulation_runner.game_state.world_map.score_cells())), 1)
        self.assertIn(grid[Location(0, 1)], self.simulation_runner.game_state.world_map.score_cells())

    def test_not_enough_score_space(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings['TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR'] = 1
        grid = self._generate_grid(1, 1)
        grid[Location(0, 0)].avatar = 'avatar'
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertEqual(len(list(self.simulation_runner.game_state.world_map.score_cells())), 0)

    def test_pickups_added(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings['TARGET_NUM_PICKUPS_PER_AVATAR'] = 1
        settings['PICKUP_SPAWN_CHANCE'] = 1
        self.simulation_runner.game_state.world_map = WorldMap(self._generate_grid(), settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertEqual(len(list(self.simulation_runner.game_state.world_map.pickup_cells())), 1)

        self.simulation_runner.update(2, self.simulation_runner.game_state)
        self.assertEqual(len(list(self.simulation_runner.game_state.world_map.pickup_cells())), 2)

    def test_pickups_applied(self):
        self.construct_simulation_runner([], [])
        grid = self._generate_grid()
        avatar = DummyAvatar()
        pickup = MockPickup(target=avatar)
        grid[Location(1, 1)].pickup = pickup
        grid[Location(1, 1)].avatar = avatar
        self.simulation_runner.game_state.world_map = WorldMap(grid, SETTINGS)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertEqual(pickup.applied_to, avatar)

    def test_pickup_spawn_chance(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings['TARGET_NUM_PICKUPS_PER_AVATAR'] = 5
        settings['PICKUP_SPAWN_CHANCE'] = 0
        grid = self._generate_grid()
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertEqual(len(list(self.simulation_runner.game_state.world_map.pickup_cells())), 0)

    @patch('simulation.pickups.pickup_types.DamageBoostPickup')
    def test_pickups_not_added_when_at_target(self, mockPickup):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings['TARGET_NUM_PICKUPS_PER_AVATAR'] = 1
        grid = self._generate_grid()
        grid[Location(0, 1)].pickup = mockPickup()
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertEqual(len(list(self.simulation_runner.game_state.world_map.pickup_cells())), 1)
        self.assertIn(grid[Location(0, 1)], self.simulation_runner.game_state.world_map.pickup_cells())

    def test_not_enough_pickup_space(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings['TARGET_NUM_PICKUPS_PER_AVATAR'] = 1
        grid = self._generate_grid(1, 1)
        grid[Location(0, 0)].generates_score = True
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertEqual(len(list(self.simulation_runner.game_state.world_map.pickup_cells())), 0)

    def test_run_turn(self):
        """
        Given:  > _
        (1)
        Expect: _ o
        """
        self.construct_simulation_runner([MoveEastDummy], [ORIGIN])
        avatar = self.get_avatar(0)

        self.assert_at(avatar, ORIGIN)
        self.run_turn()
        self.assert_at(avatar, RIGHT_OF_ORIGIN)

    def test_run_several_turns(self):
        """
        Given:  > _ _ _ _ _
        (5)
        Expect: _ _ _ _ _ o
        """
        self.construct_simulation_runner([MoveEastDummy], [ORIGIN])
        avatar = self.get_avatar(0)

        self.assertEqual(avatar.location, ORIGIN)
        [self.run_turn() for _ in range(5)]
        self.assertEqual(avatar.location, FIVE_RIGHT_OF_ORIGIN)

    def test_run_several_turns_and_avatars(self):
        """
        Given:  > _ _ _ _ _
                > _ _ _ _ _
        (5)
        Expect: _ _ _ _ _ o
                _ _ _ _ _ o
        """
        self.construct_simulation_runner([MoveEastDummy, MoveEastDummy],
                                    [ORIGIN,        ABOVE_ORIGIN])
        avatar0 = self.get_avatar(0)
        avatar1 = self.get_avatar(1)

        self.assert_at(avatar0, ORIGIN)
        self.assert_at(avatar1, ABOVE_ORIGIN)
        [self.run_turn() for _ in range(5)]
        self.assert_at(avatar0, FIVE_RIGHT_OF_ORIGIN)
        self.assert_at(avatar1, FIVE_RIGHT_OF_ORIGIN_AND_ONE_ABOVE)

    def test_move_chain_succeeds(self):
        """
        Given:  > > > > > _

        Expect: _ o o o o o
        """
        self.construct_simulation_runner([MoveEastDummy for _ in range(5)],
                                    [Location(x, 0) for x in range(5)])
        avatars = [self.get_avatar(i) for i in range(5)]

        [self.assert_at(avatars[x], Location(x, 0)) for x in range(5)]
        self.run_turn()
        [self.assert_at(avatars[x], Location(x + 1, 0)) for x in range(5)]

    def test_move_chain_fails_occupied(self):
        """
        Given:  > > x _

        Expect: x x x _
        """
        self.construct_simulation_runner([MoveEastDummy, MoveEastDummy, WaitDummy],
                                    [Location(x, 0) for x in range(3)])
        avatars = [self.get_avatar(i) for i in range(3)]

        [self.assert_at(avatars[x], Location(x, 0)) for x in range(3)]
        self.run_turn()
        [self.assert_at(avatars[x], Location(x, 0)) for x in range(3)]

    def test_move_chain_fails_occupied_by_dead_avatar(self):
        """
        Given: > > ! _

        Expect: x x ! _
        """

        self.construct_simulation_runner([MoveEastDummy, MoveEastDummy, DeadDummy], [Location(x, 0) for x in range(3)])
        avatars = [self.get_avatar(i) for i in range(3)]

        [self.assert_at(avatars[x], Location(x, 0)) for x in range(3)]
        self.run_turn()
        [self.assert_at(avatars[x], Location(x, 0)) for x in range(3)]

    def test_move_fails_collision(self):
        """
        Given: > _ <
        Expect: x _ x
        """

        self.construct_simulation_runner([MoveEastDummy, MoveWestDummy], [Location(0, 0), Location(2, 0)])
        avatars = [self.get_avatar(i) for i in range(2)]

        self.assert_at(avatars[0], Location(0, 0))
        self.assert_at(avatars[1], Location(2, 0))

        self.run_turn()

        self.assert_at(avatars[0], Location(0, 0))
        self.assert_at(avatars[1], Location(2, 0))

    def test_move_chain_fails_collision(self):
        """
        Given:  > > > _ <
        (1)
        Expect: x x x _ x
        """
        locations = [Location(0, 0), Location(1, 0), Location(2, 0), Location(4, 0)]
        self.construct_simulation_runner(
            [MoveEastDummy, MoveEastDummy, MoveEastDummy, MoveWestDummy],
            locations)
        avatars = [self.get_avatar(i) for i in range(4)]

        [self.assert_at(avatars[i], locations[i]) for i in range(4)]
        self.run_turn()
        [self.assert_at(avatars[i], locations[i]) for i in range(4)]

    def test_move_chain_fails_cycle(self):
        """
        Given:  > v
                ^ <
        (1)
        Expect: x x
                x x
        """
        locations = [Location(0, 1), Location(1, 1), Location(1, 0), Location(0, 0)]
        self.construct_simulation_runner(
            [MoveEastDummy, MoveSouthDummy, MoveWestDummy, MoveNorthDummy],
            locations)
        avatars = [self.get_avatar(i) for i in range(4)]

        [self.assert_at(avatars[i], locations[i]) for i in range(4)]
        self.run_turn()
        [self.assert_at(avatars[i], locations[i]) for i in range(4)]

    def test_move_chain_fails_spiral(self):
        """
        Given:  > > v
                  ^ <
        (1)
        Expect: x x x
                  x x
        """
        locations = [Location(0, 1),
                     Location(1, 1),
                     Location(2, 1),
                     Location(2, 0),
                     Location(1, 0)]
        self.construct_simulation_runner(
            [MoveEastDummy, MoveEastDummy, MoveSouthDummy, MoveWestDummy, MoveNorthDummy],
            locations)
        avatars = [self.get_avatar(i) for i in range(5)]

        [self.assert_at(avatars[i], locations[i]) for i in range(5)]
        self.run_turn()
        [self.assert_at(avatars[i], locations[i]) for i in range(5)]


if __name__ == '__main__':
    unittest.main()



from functools import wraps


def default_preview_user(view_func):
    @wraps(view_func)
    def wrapped(request, *args, **kwargs):
        return view_func(request, *args, **kwargs)
    return wrapped
